
/*
 * $Id: Client.cc,v 1.36 1999/09/17 23:54:06 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "CltConnMgr.h"
#include "CltOpts.h"
#include "CltXact.h"
#include "PopModel.h"
#include "ServerRep.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "ErrorMgr.h"
#include "Client.h"

#include "RobotSym.h"

#include "polyBcastChannels.h"
#include "polyErrors.h"
#include "globals.h"


ObjFarm<CltXact> Client::TheXacts;
Array<ServerRep*> Client::TheServers;
int Client:: TheShrdWorkSetCap = -1;


Client::Client(): theOriginDistr(0), theLaunchCount(0),
	thePopModel(0),
	thePublicRatio(-1), theRecurrRatio(-1), genUniqUrls(false),
	thePrivateOid(0), theWorkSetCap(-1) {

	theChannels.append(TheInfoChannel);
	startListen();
}

Client::~Client() {
}

void Client::configure(const RobotSym *cfg) {
	SockOpt opt;
	Agent::configure(cfg, opt);
	theHost.port(-1); // we do not care about port number

	configureConnMgrs(cfg, opt);

	theLaunchWin = cfg->launchWin();
	cfg->publicRatio(thePublicRatio);
	cfg->recurrRatio(theRecurrRatio);
	thePopModel = cfg->popModel();
	cfg->uniqueUrls(genUniqUrls);

	if (!thePopModel && theRecurrRatio > 0) {
		cerr << cfg->loc() << "popularity model must be specified for"
			<< " positive recurrence ratio (robot " << theKind << ')'
			<< endl;
		exit(-2);
	}	

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgrs.count());

	Agent::start();

	if (theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWorkSetFreeze) {
		Assert(theWorkSetCap < 0);
		theWorkSetCap = thePrivateOid;
		if (TheShrdWorkSetCap < 0)
			FreezeSrvWorkSetCap();
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);
	ConnMgr *mgr = x->conn()->mgr();
	Assert(mgr);
	mgr->put(x->conn());
	TheXacts.put(x);
}

void Client::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	if (theThinkDistr) {
		const double lf = TheStatPhaseMgr->loadFactor(lastLaunch);
		Assert(lf > 0);
		double delay;
		if (TheStatPhaseMgr->varyLoad()) {
			// next launch should happen [0, 2dt] from last launch
			// where dt is 1/current_rate
			static RndGen rng;
			delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
		} else {
			delay = theThinkDistr->trial()/lf;
		}
		sleepTill(lastLaunch + Time::Secd(delay));
	} else
		launch();
}

void Client::launch() {
	// select origin server to connect to
	const int sid = (int)theOriginDistr->trial();
	Assert(sid >= 0 && sid < theConnMgrs.count());

	// connect and execute transaction
	if (Connection *conn = theConnMgrs[sid]->get()) {
		theLaunchCount++;
		TheXacts.get()->exec(this, conn);
	} else {
		if (ReportError(errno))
			Comment(1) << theHost << " failed to connect to " 
				<< theConnMgrs[sid]->addr() << endc;
		Broadcast(TheXactErrChannel, (Xaction*)0);

		// if no alarms are pending, we may get stuck and submit
		// no more xactions (e.g. best effort); this is a kludge
		if (!thePendAlarmCnt)
			sleepTill(TheClock); // submit more next cycle
	}
}

void Client::configureConnMgrs(const RobotSym *cfg, const SockOpt &opt) {
	Array<NetAddr> addrs;
	if (!cfg->origins(addrs, theOriginDistr) || !addrs.count()) {
		Comment(1) << ThePrgName << ": no origin addresses specified for robot " << cfg->kind() << endc;
		FatalError(errOther);
	}
	Assert(theOriginDistr);

	// create and configure each manager
	RndDistr *d = cfg->pconnUseLmt();
	for (int i = 0; i < addrs.count(); ++i) {
		const int repIdx = GetServerRepIdx(addrs[i]);
		CltConnMgr *mgr = new CltConnMgr;
		const NetAddr &connAddr = TheCltOpts.thePxyAddr ? 
			TheCltOpts.thePxyAddr : addrs[i];
		mgr->configure(connAddr, opt, -1, d);
		mgr->originIdx(repIdx);
		theConnMgrs.append(mgr);
	}
}

void Client::genOid2(ServerRep &srvRep, OidRec &oidRec, UniqId &wid) {
	static RndGen rng;
	const bool publicOid = rng() < thePublicRatio;
	const bool repeatOid = rng() < theRecurrRatio;

	if (publicOid && srvRep.id()) {
		wid = srvRep.id(); // use server id

		// can repeat?
		if (repeatOid && srvRep.secOid() > 0) {
			oidRec.repeat = true;
			oidRec.oid = thePopModel->select(rng, srvRep.secOid(), srvRep.workSetCap());
			return;
		}

		// can ask for a new oid?
		if (srvRep.haveNewOids()) {
			oidRec.oid = srvRep.newOid();
			return;
		}

		// fall through if cannot generate a public oid
		// XXX: count these!
	}
		
	// ask for a private object
	wid = theId; // use our id

	if (repeatOid && thePrivateOid > 0) {
		oidRec.repeat = true;
		oidRec.oid = thePopModel->select(rng, thePrivateOid, theWorkSetCap);
		return;
	}

	// last resort: new private object
	oidRec.oid = ++thePrivateOid;
}

void Client::genOid(ServerRep &srvRep, OidRec &oidRec, UniqId &wid) {
	genOid2(srvRep, oidRec, wid);
	if (genUniqUrls)
		wid = UniqId::Create(); // changes every time
}

// creates new representative if needed
int Client::GetServerRepIdx(const NetAddr &addr) {
	int idx = -1;

	// check if we already know about this server
	for (int i = 0; i < TheServers.count(); ++i) {
		if (TheServers[i]->addr() == addr) {
			idx = i;
			break;
		}
	}
	
	// create rep if needed
	if (idx < 0) {
		idx = TheServers.count();
		TheServers.append(new ServerRep(addr));
	}

	Assert(TheServers[idx]);
	return idx;
}

ServerRep *Client::GetServerRep(int idx) {
	Assert(0 <= idx && idx < TheServers.count());
	return TheServers[idx];
}

void Client::FreezeSrvWorkSetCap() {
	Assert(TheShrdWorkSetCap < 0);

	TheShrdWorkSetCap = 0;
	for (int i = 0; i < TheServers.count(); ++i) {
		TheServers[i]->freezeWorkSetCap();
		TheShrdWorkSetCap += TheServers[i]->workSetCap();
	}
}
