
/*
 * $Id: ObjLifeCycle.cc,v 1.9 1999/09/07 17:46:18 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Clock.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "ObjTimes.h"
#include "ObjLifeCycle.h"

#include "PglQualifSym.h"
#include "ObjLifeCycleSym.h"

#include "gadgets.h"


ObjLifeCycle::ObjLifeCycle():
	theBday(0), theModTime(0), theExpSelector(0),
	theModVar(0), theShowModRatio(0) {
}

ObjLifeCycle::~ObjLifeCycle() {
	delete theModTime;
}

void ObjLifeCycle::configure(const ObjLifeCycleSym *cfg) {
	Assert(cfg);

	// move all the checks from 1.3 here
	bday(cfg->bday());
	cfg->variance(theModVar);
	cfg->withLmt(theShowModRatio);
	modTime(cfg->length());

	Array<QualifSym*> qs;
	cfg->expires(theExpires, theExpSelector, qs);
	// set rng
	if (theExpSelector)
		theExpSelector->rndGen(&theRng);
	for (int i = 0; i < theExpires.count(); ++i) {
		if (theExpires[i])
			theExpires[i]->rndGen(&theRng);
	}

	// interpret qualifiers
	for (int i = 0; i < qs.count(); ++i) {
		if (qs[i]) {
			Assert(qs[i]->kind() != QualifSym::qfNone);
			theExpGapTypes.append(qs[i]->kind() == QualifSym::qfLmt ? egtLmt : egtNow);
		} else
		if (theExpires[i]) {
			cerr << cfg->loc() << "expire distribution (at position " 
				<< i+1
				<< ") does not have a qualifier like `lmt' or `now'" << endl;
			exit(-2);
		} else
			theExpGapTypes.append(egtNone);
	}
}

void ObjLifeCycle::bday(RndDistr *aDistr) {
	delete theBday;
	theBday = aDistr;
	if (theBday)
		theBday->rndGen(&theRng);
}

void ObjLifeCycle::modVar(double aVar) {
	theModVar = aVar;
}

void ObjLifeCycle::modTime(RndDistr *aDistr) {
	delete theModTime;
	theModTime = aDistr;
	if (theModTime) {
		Assert(theBday);
		theModTime->rndGen(&theRng);
	}
}

// calculates modification and expiration times for a given object
// note: result times must be rounded to second resolution or
//       comparison of HTTP-equal dates may not work!
void ObjLifeCycle::calcTimes(int oid, ObjTimes &times) {
	// must have a mod time model to proceed
	if (!theModTime) {
		times.reset();
		return;
	}

	// seed random generator with object id
	// note: all distributions are linked to this rng
	theRng.seed(RndPermut(oid));

	// all time computations are in [sec]

	// used a lot
	const int now = TheClock.time().sec();

	// several "persistent" choices
	const int expIdx = theExpSelector ? (int)theExpSelector->trial() : -1;
	times.doShowLmt = theRng() < theShowModRatio;
	int bday = (int)theBday->trial(); // absolute or relative
	const int cycle = (int)theModTime->trial();

	// also persistent: expiration gap (used below)
	// the delta is persistent because that's how real servers
	// have it configured: mod + const_delta OR now + const_delta
	RndDistr *expd = expIdx >= 0 ? theExpires[expIdx] : 0;
	const int expGap = !expd ? 0 : (int)expd->trial();

	// adjust birthday: (relative "|n| secs since start" to absolute)
	if (bday <= 0)
		bday += Clock::TheStartTime.sec();

	// find lmt (without random, cycle dependent delta)
	const int cycleCnt = now > bday ? (now - bday)/cycle : 0;

	// compute lmt, adding random, cycle dependent delta (changes seed!)
	int lmt = bday + lmtDelta(oid, cycleCnt, cycle);
	if (lmt > now) // oops, random overshoot!
		lmt = bday + lmtDelta(oid, cycleCnt-1, cycle); // previous cycle

	// be sure lmt does not precede bday
	times.theLmt = Time::Sec(Max(bday, lmt));

	// calculate expiration time if needed
	if (expd) {
		if (theExpGapTypes[expIdx] == egtNow)
			times.theExp = TheClock + Time::Sec(expGap);
		else
			times.theExp = times.theLmt + Time::Sec(expGap);
	} else
		times.theExp = Time();
}

// compute lmt delta for a given cycle (with a random component)
// note: cycleCnt == number of full (completed) cycles
int ObjLifeCycle::lmtDelta(int oid, int cycleCnt, int cycleLen) {
	if (cycleCnt <= 0)
		return 0;
	theRng.seed(RndPermut(oid, cycleCnt));
	const double drift = cycleLen/2.0;
	return cycleCnt*cycleLen + (int) (theModVar * theRng(-drift, +drift));
}
