
/*
 * $Id: PolyApp.cc,v 1.22 1999/09/17 23:54:16 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <signal.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include "Socket.h"
#include "AlarmClock.h"
#include "FileScanner.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "HttpDate.h"
#include "StatCycle.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "ErrorStrTbl.h"
#include "ErrorMgr.h"
#include "Notifier.h"
#include "PolyOLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "Agent.h"
#include "UniqId.h"

#include "PglPp.h"
#include "PglParser.h"
#include "PglSemx.h"
#include "AgentSym.h"
#include "PhaseSym.h"

#include "PolyApp.h"

#include "gadgets.h"
#include "getIfAddrs.h"
#include "rndDistrs.h"
#include "polyLogTags.h"
#include "polyErrors.h"
#include "polyBcastChannels.h"
#include "shutdown.h"
#include "globals.h"


static StatCycle *TheStatCycle = 0;

inline ostream &xexit(ostream &os) { exit(-2); return os; }

// make 'rw' permissions flag to use in open(2); move this?
#ifndef S_IRWMOST
#	if defined(S_IRUSR) && defined(S_IRGRP) & defined(S_IRWXO)
#		define S_IRWMOST (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH)
#   else
#		define S_IRWMOST 0000666
#	endif
#endif

/* PolyApp */

PolyApp::PolyApp(): isIdle(false) {
	theChannels.append(TheShutdownChannel);
}

PolyApp::~PolyApp() {
	for (int i = 0; i < theLocals.count(); ++i)
		delete theLocals[i];
}

// scans all FDs once
void PolyApp::step() {
	Time tout = isIdle ? theIdleEnd - TheClock : (Time)TheOpts.theIdleTout;
	Time *toutp = &tout;

	if (TheAlarmClock.clear()) {
		if (tout < 0)
			toutp = 0;
	} else
	if (tout < 0) {
		tout = TheAlarmClock.timeLeft();
	} else {
		tout = Min(tout, TheAlarmClock.timeLeft());
	}

	const int readyCount = TheFileScanner->scan(toutp);

	if (ShutdownRequested())
		return;

	if (readyCount < 0 && errno != EINTR)
		FatalError(errno);

	// start "idle state" if needed
	if (!readyCount) {
		if (!isIdle && TheOpts.theIdleTout >= 0) {
			isIdle = true;
			theIdleBeg = TheClock;
			theIdleEnd = theIdleBeg + TheOpts.theIdleTout;
		}
	} else {
		isIdle = false;
	}

	Clock::Update();

	if (!readyCount && isIdle && TheClock.time() >= theIdleEnd)
		wasIdle(TheClock - theIdleBeg);
}

void PolyApp::begCycle(int) {
	Clock::Update(false);
	theTickCount = 0;
}

void PolyApp::endCycle() {
}

bool PolyApp::tick() {
	if (++theTickCount % 16 == 0)
		Clock::Update(false);
	return !ShutdownRequested();
}

void PolyApp::noteMsgStrEvent(BcastChannel *ch, const char *msg) {
	Assert(ch == TheShutdownChannel);
	ShutdownReason(msg);
}

void PolyApp::wasIdle(Time tout) {
	if (TheOpts.theIdleTout >= 0 && tout >= TheOpts.theIdleTout) {
		Comment << "was idle for at least " << tout << endc;
		ShutdownReason("inactivity timeout");
	}
}

bool PolyApp::handleCmdLine(int argc, char *argv[]) {

	Array<OptGrp*> opts;
	getOpts(opts);

	theCmdLine.configure(opts);
	if (!theCmdLine.parse(argc, argv))
		return false;

	// validate command line params
	for (int i = 0; i < opts.count(); ++i)
		if (!opts[i]->validate())
			return false;

	return true;
}

void PolyApp::configureLogs(int prec) {
	// redirect console output
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	// initialize log file and calculate log size
	if (TheOpts.theLogFileName) {

		// open
		int fd = ::open(TheOpts.theLogFileName, 
			O_WRONLY|O_CREAT|O_TRUNC, S_IRWMOST);
		if (fd < 0)
			cerr << "cannot create binary log; " << TheOpts.theLogFileName << ": " << error(errno) << endl << xexit;
		TheOLog.stream(TheOpts.theLogFileName, fd);

		// how often do we want to flush the log to disk
		// note: the log will not be flushed unless there is also
		// enough data to be worth flushing (e.g., I/O page size)
		TheOLog.period(Time::Sec(5));
	}

	// set log size
	if (TheOpts.theLogSize > TheOLog.capacity())
		TheOLog.resize(TheOpts.theLogSize); // we cannot shrink log size
	TheOpts.theLogSize.set(TheOLog.capacity());

	Comment.TheEchoLevel = TheOpts.theVerbLevel;

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

// get a list of all interface or fake addresses
void PolyApp::getIfaces() {
	// PglCtx::RootCtx()->report(cout, "");

	// manually configured [faked] addresses take priority
	if (TheOpts.theFakeHosts) {
		for (int i = 0; i < TheOpts.theFakeHosts.val().count(); ++i) {
			const String &s = *TheOpts.theFakeHosts.val()[i];
			in_addr a;
			a.s_addr = inet_addr(s);
			if (a.s_addr == INADDR_NONE)
				cerr << ThePrgName << ": cannot convert fake host `" << s << "' to a valid IP address" << endl << xexit;
			theIfaces.append(a);
		}
	} else
	if (!::GetIfAddrs(theIfaces))
		cerr << ThePrgName << ": cannot get a list of all available network interfaces: " << error(errno) << endl << xexit;
}

void PolyApp::makeAgents() {
	// PglCtx::RootCtx()->report(cout, "");

	getIfaces();

	// collect agents that are assigned to our host
	for (int i = 0; i < PglSemx::TheAgentsToUse.count(); ++i) {
		const AgentSym *agent = PglSemx::TheAgentsToUse[i];
		Assert(agent);

		// local if no binding or local binding
		int idx;
		if (!agent->host().addrA() || 
			theIfaces.find(agent->host().addrN(), idx))
			theLocalCfgs.append(agent);
	}

	// create and configure agents
	for (int i = 0; i < theLocalCfgs.count(); ++i)
		if (Agent *agent = makeAgent(theLocalCfgs[i]))
			theLocals.append(agent);

	if (!theLocals.count()) {
		cerr << ThePrgName << ": no agents match local host addresses" << endl;
		cerr << "agent addresses:";
		for (int i = 0; i < PglSemx::TheAgentsToUse.count(); ++i)
			cerr << ' ' << PglSemx::TheAgentsToUse[i]->host();
		cerr << endl;
		cerr << "local addresses: ";
		for (int i = 0; i < theIfaces.count(); ++i)
			cerr << ' ' << inet_ntoa(theIfaces[i]);
		cerr << endl;
		cerr << xexit;
	}
}

// configure StatPhaseMgr with user specified phases
void PolyApp::buildSchedule() {
	if (!PglSemx::TheSchedule.count()) {
		cerr << ThePrgName << ": warning: no run phases were specified; generating and using default phase" << endl;
		PhaseSym *ps = new PhaseSym();
		ps->loc(TokenLoc("-"));
		ps->name("dflt");
		PglSemx::TheSchedule.append(ps);
	}

	const StatPhase *prevPh = 0;
	for (int i = 0; i < PglSemx::TheSchedule.count(); ++i) {
		StatPhase *p = new StatPhase;
		p->configure(PglSemx::TheSchedule[i], prevPh);
		TheStatPhaseMgr.addPhase(p);
		prevPh = p;
	}
}

// top level configure routine
void PolyApp::configure() {

	configureLogs(2);

	startListen();

	TheErrorMgr.configure(polyErrorStrTbl());

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// quit nicely on some signals
	signal(SIGINT, (SignalHandler*)&ShutdownSignal);

	// set resource limits
	const int rlimit = SetMaxFD(GetMaxFD());
	if (rlimit > FD_SETSIZE)
		cerr << here << "warning: getrlimit(2) system call reports " << rlimit << " file descriptors while FD_SETSIZE #defines only " << FD_SETSIZE << "; using lower value." << endl;
	const int fdLimit = Min(FD_SETSIZE, rlimit);
	if (fdLimit <= 0)
		FatalError(errZeroFDLimit);
	TheFileScanner = TheOpts.theFileScanner.val();
	TheFileScanner->configure(fdLimit);
	TheFileScanner->ticker(this);

	// to be safe, cut about 3% from the actual limit
	Socket::TheMaxLevel = Max(0, fdLimit - 10 - fdLimit/33);
	// honor manually configured limit if it is lower
	if (TheOpts.theFDLimit >= 0 && TheOpts.theFDLimit < Socket::TheMaxLevel)
		Socket::TheMaxLevel = TheOpts.theFDLimit;
	TheOpts.theFDLimit.set(Socket::TheMaxLevel);

	// set random seeds
	RndPermutSeed(TheOpts.theRngSeed);
	RndGen::TheDefSeed = RndPermut(TheOpts.theRngSeed);

	// statistics
	TheStatCycle = new StatCycle;
	TheStatCycle->period(TheOpts.theStatCycleLen);
	buildSchedule(); // configures TheStatPhaseMgr

	// configure notification mechanism
	if (TheOpts.theNotifAddr) {
		Notifier *n = new Notifier(TheOpts.theRunLabel, TheOpts.theNotifAddr);
		if (n->active())
			TheStatCycle->notifier(n);
		else
			delete n;
	}

	atexit(&Shutdown);
}

void PolyApp::getOpts(Array<OptGrp*> &opts) {
	opts.append(&TheOpts);
}

void PolyApp::parseConfigFile(const String &fname) {
	Assert(fname);
	TheOpts.theCfgDirs.copy(PglPp::TheDirs);

	PglPp pp(fname);
	PglParser parser(&pp);

	parser.parse();
}

void PolyApp::reportCfg() {

	theCmdLine.reportRaw(Comment(1)); Comment << endc;
	theCmdLine.reportParsed(Comment(2)); Comment << endc;

	TheStatPhaseMgr.reportCfg(Comment(2) << "Phases:" << endl);
	Comment(2) << endc;

	Comment(2) << "FDs: "
		<< GetCurFD() << " out of " << GetMaxFD()
		<< " FDs can be used; safeguard limit: " << Socket::TheMaxLevel 
		<< endc;

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheOLog << bege(lgRunId) << runId << ende;

	Comment << "current time: " << TheClock.time() << " or ";
	HttpDatePrint(Comment, TheClock.time()) << endc;
	Comment << "waiting for traffic" << endc;
}


int PolyApp::run(int argc, char *argv[]) {

	ThePrgName = argv[0];

	Clock::Update();

	if (!handleCmdLine(argc, argv))
		return -1;

	parseConfigFile(TheOpts.theCfgFileName);

	Clock::Update();

	configure();
	makeAgents();

	reportCfg();

	// store errno strings
	TheOLog << bege(lgErrorStrs) << *polyErrorStrTbl() << ende;
	TheOLog.flush(); // flush prefix

	TheStatPhaseMgr.start();
	TheStatCycle->startListen(); 

	Clock::Update();

	for (int i = 0; i < theLocals.count(); ++i)
		theLocals[i]->start();

	Clock::Update();

	while (!ShutdownNow()) {
		step();
	}

	return 0;
}


#if 0 /* main() template */
int main(int argc, char *argv[]) {

	XxxApp app;
	return app.run(argc, argv);
}
#endif
