
/*
 * $Id: Server.cc,v 1.41 1999/09/17 23:54:17 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Connection.h"
#include "RndPermut.h"
#include "ContentCfg.h"
#include "Server.h"
#include "SrvConnMgr.h"
#include "ErrorMgr.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "Rnd.h"

#include "ServerSym.h"
#include "SocketSym.h"

#include "polyErrors.h"
#include "gadgets.h"
#include "globals.h"

ObjFarm<SrvXact> Server::TheXacts;

static int TheSrvHistoryWin = 1000; // hard coded; how can we estimate this?


Server::Server(): theContentDistr(0) {
	theConnMgr = new SrvConnMgr(this); // for now
}

Server::~Server() {
	delete theConnMgr;
	deaf();
}

void Server::configure(const ServerSym *cfg, bool bind2iface) {
	Assert(theConnMgr);

	SockOpt opt;
	Agent::configure(cfg, opt);

	configureContents(cfg);

	theHistory.window(TheSrvHistoryWin);

	// setup listen socket

	Must(theSock.create());
	Must(theSock.blocking(false));
	Should(theSock.reuseAddr(true));
	Must(theSock.configure(opt));

	if (bind2iface)
		Must(theSock.bind(theHost));
	else
		Must(theSock.bind(NetAddr(0, theHost.port())));

	Must(theSock.listen());

	int acceptLmt = -1;
	cfg->acceptLmt(acceptLmt);
	theConnMgr->configure(theHost, opt, cfg->pconnUseLmt());
	theConnMgr->acceptLmt(acceptLmt);

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Server::start() {
	Assert(theSock);
	Assert(theConnMgr);
	Agent::start();
	theReserv = TheFileScanner->setFD(theSock.fd(), dirRead, this);
}

void Server::noteXactDone(SrvXact *x) {
	Assert(x);
	theConnMgr->put(x->conn());
	TheXacts.put(x);
}

void Server::noteReadReady(int) {
	if (!theConnMgr->accept(theSock)) {	// fatal error
		deaf();
		FatalError(errServerGone);
	}
}

void Server::noteConnReady(Connection *conn) {
	startXact(conn); // that's it?!
}

void Server::deaf() {
	if (theSock) {
		Comment << "server " << theHost << " is closing listen socket " << theSock.fd() << endc;
		theSock.close();
	}
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
}

void Server::startXact(Connection *conn) {
	SrvXact *x = TheXacts.get();
	x->exec(this, conn,
		theThinkDistr ? Time::Secd(theThinkDistr->trial()) : Time());
}

ContentCfg *Server::selectContent(int oid) {
	theContentDistr->rndGen()->seed(RndPermut(oid));
	const int cid = (int)theContentDistr->trial();
	Assert(cid >= 0 && cid < theContents.count());
	return theContents[cid];
}

void Server::configureContents(const ServerSym *cfg) {
	Array<ContentSym*> ccfgs;
	if (!cfg->contents(ccfgs, theContentDistr) || !ccfgs.count()) {
		Comment(0) << ThePrgName << ": no content specs for server " << cfg->kind() << endc;
		FatalError(errOther);
	}
	Assert(theContentDistr);

	// create and configure each content cfg
	for (int i = 0; i < ccfgs.count(); ++i) {
		Assert(ccfgs[i]);
		ContentCfg *cc = new ContentCfg;
		cc->configure(ccfgs[i]);
		Check(cc->theSize);
		theContents.append(cc);
	}
}
