
/*
 * $Id: PglSizeSym.cc,v 1.6 1999/09/07 17:47:49 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "pgl.h"

#include <math.h>

#include "PglNumSym.h"
#include "PglSizeSym.h"


SizeSym::SizeSym(const BigSize &aVal): ExpressionSym("size"), theVal(aVal) {
}

bool SizeSym::isA(const String &type) const {
	return ExpressionSym::isA(type) ||
		type == "Size" || type == "size";
}


SynSym *SizeSym::clone(const String &type) const {
	if (isA(type))
		return new SizeSym(theVal);
	return ExpressionSym::clone(type);
}

ExpressionSym *SizeSym::unOper(const String &op) const {
	if (*op == '+')
		return new SizeSym(+theVal);
	else
	if (*op == '-')
		return new SizeSym(-theVal);
	else
		return ExpressionSym::unOper(op);
}


ExpressionSym *SizeSym::bnOper(const String &op, const SynSym &s) const {
	if (op.len() != 1)
		return ExpressionSym::bnOper(op, s);

	// division is defined for both Size and Num
	const bool forceSS = *op == '/' && s.canBe("size");

	if (*op == '+' || *op == '-' || forceSS)
		return operSS(op, s); // size op size

	if (*op == '*' || *op == '/')
		return operSN(op, s); // size op num
	
	return ExpressionSym::bnOper(op, s);
}

ExpressionSym *SizeSym::operSS(const String &op, const SynSym &s) const {
	BigSize otherVal;
	
	if (s.isA("size")) {
		otherVal = ((const SizeSym&)s.cast("size")).val();
	} else {
		SizeSym *ts = (SizeSym*)s.clone("size");
		if (!ts)
			return ExpressionSym::bnOper(op, s);
		otherVal = ts->val();
		delete ts;
	}

	if (*op == '/' && otherVal == 0)
		cerr << theLoc << "denominator is zero" << endl << xexit;

	switch (*op) {
		case '+': return new SizeSym(theVal + otherVal);
		case '-': return new SizeSym(theVal - otherVal);
		//case '*': return new SizeSym(theVal * otherVal);
		case '/': return new NumSym(theVal / otherVal);
		default: return ExpressionSym::bnOper(op, s);
	}
}

ExpressionSym *SizeSym::operSN(const String &op, const SynSym &s) const {
	double otherVal = 0;
	
	if (s.isA("float")) {
		otherVal = ((const NumSym&)s.cast("float")).val();
	} else {
		NumSym *fs = (NumSym*)s.clone("float");
		if (!fs)
			return ExpressionSym::bnOper(op, s);
		otherVal = fs->val();
		delete fs;
	}

	if (*op == '/' && fabs(otherVal) < 1e-8)
		cerr << theLoc << "denominator is "
			<< (otherVal ? "[virtually]" : "")
			<< " zero" << endl << xexit;

	switch (*op) {
		case '*': return new SizeSym(theVal * otherVal);
		case '/': return new SizeSym(theVal * otherVal);
		default: return ExpressionSym::bnOper(op, s);
	}
}

ostream &SizeSym::print(ostream &os, const String &) const {
	return os << theVal;
}
