
/*
 * $Id: Socket.cc,v 1.11 1999/09/07 17:48:29 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "xstd.h"

#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>  /* for SIOCGIFCONF on Solaris */
#endif
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/tcp.h> /* for TCP_NODELAY */
#include <net/if.h>      /* for ifconf */
#include <arpa/inet.h>
#include <netdb.h>

#include "Socket.h"
#include "Assert.h"

bool Socket::IsEnabled = true;
int  Socket::TheMaxLevel = -1;
int  Socket::TheLevel = 0;

bool Socket::create() {
	return create(PF_INET, SOCK_STREAM, 0);
}

bool Socket::create(int domain, int type, int protocol) {

	if (TheMaxLevel >= 0 && TheLevel >= TheMaxLevel) {
		errno = EMFILE;
		return false;
	}
	
	theFD = ::socket(domain, type, protocol);

	if (theFD >= 0)
		TheLevel++;

	// most MaxLevel users assume that MaxLevel means max FD!
	if (TheMaxLevel >= 0 && theFD >= TheMaxLevel) {
		close();
		errno = EMFILE;
		return false;
	}

	return theFD >= 0;
}

bool Socket::bind(const NetAddr &addr) {
	static sockaddr_in S;

	S.sin_family = AF_INET;
	S.sin_port = htons(addr.port());
	if (addr.addrA())
		S.sin_addr = addr.addrN();
	else
		S.sin_addr.s_addr = htonl(INADDR_ANY);

	return !IsEnabled ||
		::bind(theFD, (sockaddr *) &S, sizeof(S)) == 0;
}

bool Socket::connect(const NetAddr &addr) {
	static sockaddr_in s;

	s.sin_family = AF_INET;
	s.sin_addr = addr.addrN();
	s.sin_port = htons(addr.port());

	if (IsEnabled && ::connect(theFD, (sockaddr *) &s, sizeof(s)) < 0)
		return errno == EINPROGRESS;
	return true;
}

bool Socket::listen() {
	return !IsEnabled ||
		::listen(theFD, FD_SETSIZE) == 0;
}

Socket Socket::accept(sockaddr *addr_p, socklen_t *addr_len_p) {

	if (TheMaxLevel >= 0 && TheLevel >= TheMaxLevel) {
		errno = EMFILE;
		return -1;
	}

	sockaddr addr;
	socklen_t addr_len = sizeof(addr);
	if (!addr_p) {
		addr_p = &addr;
		addr_len_p = &addr_len;
	} else
		Assert(addr_len_p);

	const int fd = IsEnabled ?
		::accept(theFD, addr_p, addr_len_p) :
		(errno = EWOULDBLOCK, -1);

	if (fd >= 0)
		TheLevel++;

	return fd;
}

Size Socket::read(void *buf, Size sz) {
	return IsEnabled && sz > 0 ? (Size)::read(theFD, buf, sz) : (Size)0;
}

Size Socket::write(const void *buf, Size sz) {
	return IsEnabled && sz > 0 ? (Size)::write(theFD, buf, sz) : sz;
}

bool Socket::close() {
	if (theFD >= 0) {
		const int res = ::close(theFD);
		theFD = -1;
		TheLevel--;
		return res == 0;
	}
	errno = EBADF;
	return false;
}

bool Socket::sendBuf(Size sz) {
#ifdef SO_SNDBUF
	return setOpt(SO_SNDBUF, (const void*)&sz, sizeof(sz));
#else
	sz = 0; // make compiler happy
	return false;
#endif
}

bool Socket::recvBuf(Size sz) {
#ifdef SO_RCVBUF
	return setOpt(SO_RCVBUF, (const void*)&sz, sizeof(sz));
#else
	sz = 0; // make compiler happy
	return false;
#endif
}

bool Socket::getIfConf(struct ifconf &cfg) {
#ifdef SIOCGIFCONF
	return ioctl(theFD, SIOCGIFCONF, &cfg) >= 0;
#else
	errno = EINVAL;
	return false;
#endif
}

bool Socket::configure(const SockOpt &opt) {
	if (opt.nagle && !tcpDelay(opt.nagle > 0))
		return false;

	if (opt.linger >= 0 && !linger(opt.linger))
		return false;

	return true;
}

int Socket::flags() const {

	const int flags = fcntl(theFD, F_GETFL);
	return flags;
}

/* local port */
int Socket::lport() const {
	sockaddr_in S;
	socklen_t len = sizeof(S);

	if (getsockname(theFD, (sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

/* remote (peer) port */
int Socket::rport() const {
	sockaddr_in S;
	socklen_t len = sizeof(S);

	if (getpeername(theFD, (sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

Time Socket::rcvTimeout() const {
#ifdef SO_RCVTIMEO
	Time t;
	socklen_t len = sizeof(t);
	if (getOpt(SO_RCVTIMEO, (void*)&t, &len))
		return t;
#endif
	return Time();
}

Size Socket::sendBuf() const {
#ifdef SO_SNDBUF
	Size sz;
	socklen_t len = sizeof(sz);
	if (getOpt(SO_SNDBUF, (void*)&sz, &len))
		return sz;
#endif
	return Size();
}

Size Socket::recvBuf() const {
#ifdef SO_RCVBUF
	Size sz;
	socklen_t len = sizeof(sz);
	if (getOpt(SO_RCVBUF, (void*)&sz, &len))
		return sz;
#endif
	return Size();
}

/*
 * Note: Solaris 2.4's socket emulation doesn't allow you
 * to determine the error from a failed non-blocking
 * connect and just returns EPIPE. -- fenner@parc.xerox.com
 */
int Socket::errNo() const {
	int e;
	socklen_t len = sizeof(e);

	if (getOpt(SO_ERROR, (void*)&e, &len))
		return errno = e;
	else
		return -1;
}

bool Socket::peerAddr(sockaddr *S, socklen_t &len) const {
	return getpeername(theFD, S, &len) == 0;
}

NetAddr Socket::peerAddr() const {
	sockaddr_in s;
	socklen_t len = sizeof(s);
	if (peerAddr((sockaddr *) &s, len))
		return NetAddr(inet_ntoa(s.sin_addr), ntohs(s.sin_port));
	else
		return NetAddr();
}

bool Socket::reuseAddr(bool set) {
#ifdef SO_REUSEADDR
	return setOpt(SO_REUSEADDR, set);
#else
	return set = false; // make compiler happy
#endif
}

bool Socket::linger(Time delay) {
#ifdef SO_LINGER
	static struct linger l = { delay.sec() > 0, delay.sec() };
	return setOpt(SO_LINGER, &l, sizeof(l));
#else
	return false && sizeof(delay); // make compiler happy
#endif
}

bool Socket::blocking(bool set) {

	int flags = this->flags();

	if (set)
		flags &= ~O_NONBLOCK;
	else
		flags |=  O_NONBLOCK;

	return fcntl(theFD, F_SETFL, flags) != -1;
}

bool Socket::reusePort(bool set) {
#ifdef SO_REUSEPORT
	return setOpt(SO_REUSEPORT, set);
#else
	return set = false; // make compiler happy
#endif
}

bool Socket::tcpDelay(bool set) {
#ifdef TCP_NODELAY
	int optVal = (int)!set;
	// must cast to char* on Solaris
	return ::setsockopt(theFD, IPPROTO_TCP, TCP_NODELAY, (const char*)&optVal, sizeof(optVal)) == 0;
#else
	return set = false; // make compiler happy
#endif
}

bool Socket::isBlocking() const {
	return flags() & O_NONBLOCK == 0;
}


bool Socket::setOpt(int optName, bool set) {
	int optVal = (int)set;
	return setOpt(optName, &optVal, sizeof(optVal));
}

bool Socket::setOpt(int optName, const void *optVal, socklen_t optLen) {
	// must cast to char* on Solaris
	return ::setsockopt(theFD, SOL_SOCKET, optName, (const char*)optVal, optLen) == 0;
}

bool Socket::getOpt(int optName, void *optVal, socklen_t *optLen) const {
	// must cast to char* on Solaris
	return ::getsockopt(theFD, SOL_SOCKET, optName, (char*)optVal, optLen) == 0;
}
